;******************************************************************* 
;* INVERT7.ASM   PIC PWM DC->AC Inverter Controller 
;*         Generates 60Hz PWM waveform for primary of inverter Xfrm 
;*         Begins at 0.5 max. width and uses ADC to sample average 
;*         AC output voltage.  Also controls main power relays and 
;*         corrects load power factor
;* 
;* M. Csele  original version: 96/12/04  version 7: 98/12  
;* 
;* Hardware: Adapted for PIC Applications PCB with on-board ADC0831
;*
;*           RA0 = ADC DATA Line
;*           RA1 = ADC CLK Line
;*           RA2 = ADC CS Line
;*           RA3 = Alarm Beeper Output 
;*                 (Used to announce startup and shutdown)  
;*           RA4 = [FUTURE OPTION] Shutdown Input (Active Low)  
;*
;*           RB0 = [FUTURE OPTION] Serial Receive/Transmit 
;*           RB1 = Power Factor Correction capacitor relay
;*                 (Switches in PF capacitor above 500VA load)
;*           RB2 = Error LED Output 
;*                 (Error code when shutdown due to fault) 
;*           RB3 = OverLoad LED Output
;*                 (Fast Blink = permissive overload)
;*                 (Constant   = Vout < 115V at max pulse width) 
;*                 (Slow Blink = shutdown after 6 secs of overload) 
;*           RB4 = PWM Change Status LED Output 
;*                 (Lights when PW changing due to voltage error > 2) 
;*           RB5 = Main driver bank Relay ON Output 
;*                 (Controls main power relays for xfmrs)     
;*           RB6 = Phase 2 Output to driver transistors (ACTIVE LOW) 
;*           RB7 = Phase 1 Output to driver transistors (ACTIVE LOW) 
;******************************************************************* 

        LIST    p=16C84       ;PIC16C84 is the target processor
        INCLUDE "P16C84.INC"  ;Include file with register defines

;Register Defines
Temp       equ    2F       ;temp storage during misc functions
ADCresult  equ    2E       ;stores ADC reading
ADCCtr     equ    2D       ;stores count for reading 8 bits from serial ADC
PulseWidth equ    2C       ;Width of var. part of pulse in 10uS increments
Timer      equ    2B       ;Used as a count-down timer
ErrorValue    equ   2A
ErrorTimerLo  equ   29
ErrorTimerHi  equ   28
LedBlinkTimer equ   27     ;Used for blinking LED during max. load     

;PORT A Defines
ADCCLK    equ    1
ADCDAT    equ    0
ADCCS     equ    2
BEEP      equ    3

;PORT B Defines
PHASE1    equ    7
PHASE2    equ    6
PWRRELAY  equ    5
LEDPWM    equ    4
LEDOVERLD equ    3
LEDERROR  equ    2
PFCRELAY  equ    1

Main 
        movlw   b'00000001'   ;Port B all outputs except unused RB0
        tris    PORTB
        movlw   b'11000000'   ;All Outputs LOW except drivers OFF (high)
        movwf   PORTB

        movlw   b'00010001'   ;Port A all inputs except RA4 and RA0 (DATA)
        tris    PORTA
        movlw   b'00000000'   ;All Outputs LOW
        movwf   PORTA

        movlw   .0          ;Initialize Pulse Width to 0.5 Max. (Min)
        movwf   PulseWidth

        movlw   b'10000110'   ;Set timer 0 to overflow at 240Hz.  7200Hz in =
                              ;Internal clock, Divide by 128 prescaler
        bsf     STATUS,RP0    ;select Page 1
        movwf   OPTION_REG
        bcf     STATUS,RP0    ;select Page 0
        clrwdt

        movlw   .226          ;Initialize timer for 240Hz overflow
        movwf   TMR0

        bsf     PORTA,BEEP    ;Beep for 1 second
        movlw   .240          ;1 second delay using timer
        movwf   Timer
NextDly
        call    WaitForTmr
        decfsz  Timer,f
        goto    NextDly
        bcf     PORTA,BEEP 

        bsf     PORTB,PHASE1   ;Turn Phases OFF
        bsf     PORTB,PHASE2
        bsf     PORTB,PWRRELAY ;Main Power Relay ON
        bcf     PORTB,PFCRELAY ;Power Factor Correction OFF
        
        ;Starting Value: Minimum pulse width
        movlw   .1
        movwf   PulseWidth

;** Inverter main loop: Generate 60Hz PWMed Sine Wave and check system 
MainLoop
        ;Phase A half-cycle output 
        ;Generate fixed width pulse of 1/240th sec.
        bcf     PORTB,PHASE1  ;Phase A ON (active low)

        ;During Phase A fixed delay, check output voltage and adjust
        ;Pulse Width Accordingly for the next cycle
        call    ReadADC 
        call    AdjustPulseWidth

        call    WaitForTmr    ;Wait for 1/240th second to elapse

        ;Generate variable width pulse of 0 to 1/240th sec
        movfw   PulseWidth
        movwf   Timer
NextPhaseADly
        call    Delay10uS     ;Generate from 1 to 255 10uS delays
        decfsz  Timer,f
        goto    NextPhaseADly    

        bsf     PORTB,PHASE1  ;Phase A OFF 
        call    WaitForTmr    ;Wait for rest of 1/240th sec period

        ;Phase B half-cycle output 
        ;Generate fixed width pulse of 1/240th sec.
        bcf     PORTB,PHASE2  ;Phase B ON (active low)
        call    WaitForTmr    ;Wait for 1/240th second to elapse

        ;Generate variable width pulse of 0 to 1/240th sec
        movfw   PulseWidth
        movwf   Timer
NextPhaseBDly
        call    Delay10uS     ;Generate from 1 to 255 10uS delays
        decfsz  Timer,f
        goto    NextPhaseBDly    

        bsf     PORTB,PHASE2  ;Phase B OFF 
        call    WaitForTmr    ;Wait for rest of 1/240th sec period

        goto    MainLoop


;******************************   
;** Called Functions Library **
;******************************   
;** 10 microsecond Delay Function ... includes 4 cycles for call/return     
Delay10uS
        nop       ;6uS+4uS for Call/Return = 10uS total
        nop 
        nop
        nop
        nop
        nop
        nop       ;ADDED to increase max PW to 8.06 mS (This delay=11.9uS)
        return

;** Function to wait for RTCC to expire and re-load it for 240Hz
WaitForTmr 
        movf   TMR0,w
        btfss  STATUS,Z
        goto   WaitForTmr
;Set RTCC for divide by 30 again 
        movlw  .226
        movwf  TMR0
        return

;** Function to read one byte from the ADC and store it in ADCresult
;   Result is a number representing Output Voltage (AC).  Range is 0 to 255
;   Target output (binary at the ADC) is 127 (= 120 VAC)
ReadADC                             
	 movlw   0            ; Clear out ADCresult register for new data
	 movwf   ADCresult
	 bcf     PORTA,ADCCS  ; Activate /CS line on the ADC to start
	 movlw   9            ; Set number of bits to read in = 8 +1 start
	 movwf   ADCCtr       ; Load this into the counter register
	 bsf     PORTA,ADCCLK ; Pulse the ADC clock line High then Low
	 nop                  ;   signals the start of a conversion
	 bcf     PORTA,ADCCLK

Get_One_Loop                  ; This is the main loop, should execute 9 times
	 bsf     PORTA,ADCCLK ; Pulse the ADC clock line hight then low
	 nop                  ;  to get the first bit
	 bcf     PORTA,ADCCLK
	 btfsc   PORTA,ADCDAT ; Check the value of the data bit
	 goto    SetCarry
	 bcf     STATUS,C     ; If its a zero, put a zero into the Carry
	 goto    RotateBits
SetCarry
	 bsf     STATUS,C     ; Otherwise its a 1, put that into the carry
RotateBits
	 rlf     ADCresult    ; Shift bits left from carry into ADReslt
	 decfsz  ADCCtr       ; Decrement Counter -
	 goto    Get_One_Loop ;   Do it 9 times
	 bsf     PORTA,ADCCS  ; When done, Disable the ADC
         return

;** Function to adjust Pulse Width
;   PulseWidth must be constrained to a range of 1 to 255
;If required for stability, ERROR could be divided by 2 using a shift

AdjustPulseWidth
        ;First, calculate error.  Error = 127(target)-ADCresult(actual)
        ;Error will be a 2's complement number in the range -128 to +127
        movfw  ADCresult
        sublw  .127           ;Calculate error (w = 127-Vactual)
        movwf  ErrorValue

        ;If error is a POSITIVE quantity, Vactual is too low ... PW++
        ;If error is a NEGATIVE quantity, Vactual is too high ... PW--
        ;If error is zero, we are at target output voltage ... do not adjust
        btfsc  STATUS,Z
        goto   ZeroError

        ;Error is a 2's complement number (0x00-0x7F=pos, 0xFF-0x80=neg)
        bsf    PORTB,LEDPWM   ;Light PWM Change LED (yellow)
        btfsc  ErrorValue,7   ;If High Bit is set, it is negative
        goto   NegativeError
 
PositiveError
        ;Output voltage is too low, increase PulseWidth
        ;Since error is a positive quantity (0x00 to 0x7F), add it to PW
        movfw  ErrorValue
        addwf  PulseWidth,f   ;PW = PW+Error

        ;Now check if we overflowed (ie PulseWidth > 255: Carry set)
        btfss  STATUS,C
        goto   AdjustPowerFactor

        movlw  .255            ;Maximum pulse width = 255
        movwf  PulseWidth

        ;PERMISSIVE OVERLOAD LOGIC: If the inverter is at max. PW and the
        ;error is small (ie output voltage > 105 VAC), allow it to run.
        ;If the error is large, it is a true overload so time 10 seconds
        ;then shutdown if it persists.

        movfw   ErrorValue
        sublw   .20             ;Check if error is > .20

        btfss   STATUS,C        ;C=1 if Error < .20 
        goto    LargeError

        incf    LedBlinkTimer,f ;Blink the Overload LED quickly

        bcf     PORTB,LEDOVERLD
        btfsc   LedBlinkTimer,4 ;Blink Overload Led at 2Hz to show max. output
        bsf     PORTB,LEDOVERLD
        goto    AdjustPowerFactor

LargeError
        ;The error is large so this is a true overload condition
        bsf    PORTB,LEDOVERLD ;Show that an overload is occurring (Max PW)

        incf   ErrorTimerLo   ;Increment Overload timer at 60Hz
        btfss  STATUS,Z       ;to keep track of time
        goto   TestErrorTime
        incf   ErrorTimerHi   ;High counter  

TestErrorTime
        movfw  ErrorTimerHi   ;Test for 10 seconds of overload error
        sublw  .2             ;w=2-w  Test for high Byte = 2
        btfss  STATUS,Z
        goto   AdjustPowerFactor
        movfw  ErrorTimerLo
        sublw  .88            ;w=88-w Test low byte = 88 (600 ticks=10 sec)
        btfss  STATUS,Z
        goto   AdjustPowerFactor

        ;Overload has persisted > 10 seconds.  Shutdown and Blink Error Code        
        bsf    PORTB,PHASE1  ;Shut down Phases
        bsf    PORTB,PHASE2
        bcf    PORTB,PWRRELAY;Main relay off
        
ErrorLoop
        bsf     PORTA,BEEP    ;Beep for 0.25 second
        bsf     PORTB,LEDERROR  ;Light Error LED
        bcf     PORTB,LEDOVERLD ;Light Overload LED alternately 
        movlw   .60           ;0.25 second delay using timer
        movwf   Timer
NextDlyA
        call    WaitForTmr
        decfsz  Timer,f
        goto    NextDlyA
        bcf     PORTA,BEEP 
        bcf     PORTB,LEDERROR 
        bsf     PORTB,LEDOVERLD

        movlw   .60           ;0.25 second delay using timer
        movwf   Timer
NextDlyB
        call    WaitForTmr
        decfsz  Timer,f
        goto    NextDlyB

        goto    ErrorLoop       ;Never ending loop


ZeroError
        bcf    PORTB,LEDPWM   ;Turn off Overload and PWM change LEDs
        bcf    PORTB,LEDOVERLD
        clrf   ErrorTimerLo   ;Clear timers for overload error
        clrf   ErrorTimerHi
        goto   AdjustPowerFactor   ;Error is zero, so no adjustment needed

NegativeError
        ;Output voltage too high ... decrease pulse width
        ;Error must be a 2's Comp. negative quantity (0xFF to 0x80)
        bcf    PORTB,LEDOVERLD ;Turn off overload LED if still on
        clrf   ErrorTimerLo   ;Clear timers for overload error
        clrf   ErrorTimerHi

        ;Convert Error to a magnitude (0x00 to 0x7F) and compare it to PW
        ;If magnitude of error > PW, set PW to minimum, otherwise, add 
        movfw  ErrorValue
        comf   ErrorValue     ;Invert Error then increment to convert ...
        incf   ErrorValue,f   ;... Signed number into a positive integer
        
        ;Subtract PW-Error then Check if magnitude of error > PW
        movfw  ErrorValue
        subwf  PulseWidth,f     ;PW=PW-|ErrorValue|

        
        btfss  STATUS,C         ;C=0 if result is negative (ie Error>PW)
        goto   SetMinPulseWidth

        btfss  STATUS,Z         ;If result is zero, set to min width
        goto   AdjustPowerFactor

SetMinPulseWidth
        movlw  .1               ;Set PW = 1 (mimimum width)
        movwf  PulseWidth

;Before exiting AdjustPulseWidth, check if load is > 500VA.  If so, switch
;in power factor correction.  Not perfect, but helps immensely with motor
;starting!  On large resistive loads it simply increases current consumption.
;This logic is required to prevent an unloaded inverter from 'seeing' a huge
;capacitive load.
AdjustPowerFactor
        movfw   PulseWidth
        sublw   .55             ;Check if PW > 5.0 mS

        btfss   STATUS,C        ;C=1 if PW < 5.0 mS (small load)
        goto    LargeLoad

        bcf     PORTB,PFCRELAY  ;Swich out PF correction
        return

LargeLoad
        bsf     PORTB,PFCRELAY  ;Switch in capacitor to correct PF
        return


       END
